from dataclasses import asdict
from importlib import resources

from jinja2 import Template

from thriftpyi.entities import Content


def render(content: Content) -> str:
    template = Template(resources.read_text("thriftpyi.templates", "stub.html"))
    return template.render(**asdict(content))  # type: ignore



from dataclasses import dataclass, field
from typing import List, Optional


@dataclass
class Field:
    name: str
    type: str
    value: Optional[str] = None


@dataclass
class Method:
    name: str
    args: List[Field]
    return_type: Optional[str]


@dataclass
class Service:
    name: str
    methods: List[Method]


@dataclass
class Class:
    name: str
    fields: List[Field]


@dataclass
class Content:
    imports: List[str] = field(default_factory=list)
    errors: List[Class] = field(default_factory=list)
    enums: List[Class] = field(default_factory=list)
    structs: List[Class] = field(default_factory=list)
    services: List[Service] = field(default_factory=list)
    is_async: bool = False



import os
from pathlib import Path
from typing import List


def save(data: str, to: str) -> None:
    Path.mkdir(Path(to).parent, exist_ok=True)
    with open(to, "w+", encoding="utf-8") as f:
        f.write(data)


def list_interfaces(interfaces_dir: str) -> List[str]:
    interfaces = []
    for root, _, files in os.walk(interfaces_dir):
        interfaces.extend(
            [os.path.join(root, file) for file in files if file.endswith(".thrift")]
        )
    return interfaces


def build_output_path(interface: str, output_dir: str) -> str:
    return os.path.join(output_dir, Path(interface).name.replace(".thrift", ".pyi"))


def get_name(path: str) -> str:
    return Path(path).name[: -len(Path(path).suffix)]



from types import ModuleType
from typing import Dict, List, Optional, cast

import thriftpy2
from thriftpy2.thrift import TPayloadMeta

from thriftpyi.utils import get_python_type


class InterfaceProxy:
    def __init__(self, interface_path: str):
        self.module = thriftpy2.load(interface_path)

    def get_services(self) -> List["ServiceProxy"]:
        return [
            ServiceProxy(member)
            for member in self.module.__dict__.values()
            if hasattr(member, "thrift_services")
        ]

    def get_imports(self) -> Dict[str, ModuleType]:
        return {
            name: item
            for name, item in self.module.__dict__.items()
            if isinstance(item, ModuleType)
        }

    def get_errors(self) -> List["ClassProxy"]:
        return [
            ClassProxy(member)
            for name, member in self.module.__dict__.items()
            if isinstance(member, TPayloadMeta) and hasattr(member, "args")
        ]

    def get_enums(self) -> List["EnumProxy"]:
        return [
            EnumProxy(member)
            for name, member in self.module.__dict__.items()
            if hasattr(member, "_NAMES_TO_VALUES")
        ]

    def get_structs(self) -> List["ClassProxy"]:
        return [
            ClassProxy(member)
            for name, member in self.module.__dict__.items()
            if isinstance(member, TPayloadMeta) and not hasattr(member, "args")
        ]


class ClassProxy:
    def __init__(self, tclass: TPayloadMeta):
        self._tclass = tclass
        self.name = tclass.__name__
        self.module_name = tclass.__module__

    def get_fields(self) -> List["FieldProxy"]:
        default_spec = dict(self._tclass.default_spec)
        return [
            FieldProxy(thrift_spec, default_spec=default_spec)
            for thrift_spec in self._tclass.thrift_spec.values()
        ]


class EnumProxy(ClassProxy):
    def get_fields(self) -> List["FieldProxy"]:
        fields = self._tclass._NAMES_TO_VALUES  # pylint: disable=protected-access
        ttype = self._tclass._ttype  # pylint: disable=protected-access
        return [
            FieldProxy((ttype, name, True), default_spec={name: value})
            for name, value in fields.items()
        ]


class ServiceProxy:
    def __init__(self, service):
        self.service = service
        self.name = service.__name__
        self.module_name = service.__module__

    def get_methods(self) -> List[str]:
        return cast(List[str], self.service.thrift_services[:])

    def get_args_for(self, method_name) -> List["VarProxy"]:
        method_args = getattr(self.service, f"{method_name}_args").thrift_spec.values()
        return [VarProxy(arg) for arg in method_args]

    def get_return_type_for(self, method_name) -> str:
        returns = getattr(self.service, f"{method_name}_result").thrift_spec.get(0)
        if returns is None:
            return "None"
        return VarProxy(returns).reveal_type_for(self.module_name)


class VarProxy:
    def __init__(self, thrift_spec: tuple):
        ttype, name, *meta, _ = thrift_spec
        self._ttype: int = ttype
        self.name: str = name
        self._meta = meta
        self._is_required: bool = True

    def reveal_type_for(self, module_name: str, strict_optional: bool = True) -> str:
        pytype = get_python_type(
            self._ttype, self._is_required and strict_optional, self._meta
        )
        start, _, end = pytype.rpartition(f"{module_name}.")
        return start + end


class FieldProxy(VarProxy):
    def __init__(self, thrift_spec: tuple, default_spec: Dict[str, str]):
        super().__init__(thrift_spec)
        self._default_value = default_spec[self.name]
        self._is_required = thrift_spec[-1]

    def reveal_value(self, strict_optional: bool = True) -> Optional[str]:
        if (
            self._default_value is not None
            or not self._is_required
            or not strict_optional
        ):
            if isinstance(self._default_value, str):
                return f'"{self._default_value}"'
            return f"{self._default_value}"
        return None



from typing import List, Tuple

from thriftpy2.thrift import TType


def get_python_type(ttype: int, is_required: bool, meta: List) -> str:
    type_map = {
        TType.BOOL: _get_bool,
        TType.DOUBLE: _get_double,
        TType.BYTE: _get_byte,
        TType.I16: _get_i16,
        TType.I32: _get_i32,
        TType.I64: _get_i64,
        TType.STRING: _get_str,
        TType.STRUCT: _get_struct,
        TType.MAP: _get_map,
        TType.SET: _get_set,
        TType.LIST: _get_list,
    }
    pytype = type_map[ttype](meta)
    if not is_required:
        pytype = f"Optional[{pytype}]"
    return pytype


def _get_bool(meta: List) -> str:
    del meta
    return "bool"


def _get_double(meta: List) -> str:
    del meta
    return "float"


def _get_byte(meta: List) -> str:
    del meta
    return "int"


def _get_binary(meta: List) -> str:
    del meta
    if TType.BINARY == TType.STRING:
        return "str"
    return "bytes"


def _get_i16(meta: List) -> str:
    del meta
    return "int"


def _get_i32(meta: List) -> str:
    del meta
    return "int"


def _get_i64(meta: List) -> str:
    del meta
    return "int"


def _get_str(meta: List) -> str:
    del meta
    return "str"


def _get_struct(meta: List) -> str:
    return f"{meta[0].__module__}.{meta[0].__name__}"


def _get_list(meta: List) -> str:
    subttype, submeta = _unpack_meta(meta)
    return f"List[{get_python_type(subttype, True, submeta)}]"


def _get_map(meta: List) -> str:
    key, value = meta[0]
    key_ttype, key_meta = _unpack_meta([key])
    value_ttype, value_meta = _unpack_meta([value])
    key_pytype = get_python_type(key_ttype, True, key_meta)
    value_pytype = get_python_type(value_ttype, True, value_meta)
    return f"Dict[{key_pytype}, {value_pytype}]"


def _get_set(meta: List) -> str:
    subttype, submeta = _unpack_meta(meta)
    return f"Set[{get_python_type(subttype, True, submeta)}]"


def _unpack_meta(meta: List) -> Tuple[int, List]:
    try:
        subttype, submeta = meta[0]
    except TypeError:
        subttype, submeta = meta[0], None
    return subttype, [submeta]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['thriftpyi', 'thriftpyi.templates']

package_data = \
{'': ['*']}

install_requires = \
['autoflake>=1.3,<2.0',
 'black>=19.3b0,<20.0',
 'click>=7.0,<8.0',
 'jinja2>=2.10,<3.0',
 'thriftpy2>=0.4.2,<0.5.0']

entry_points = \
{'console_scripts': ['thriftpyi = thriftpyi.cli:main']}

setup_kwargs = {
    'name': 'thrift-pyi',
    'version': '0.3.0',
    'description': 'This is simple `.pyi` stubs generator from thrift interfaces',
    'long_description': '========\nOverview\n========\n\n.. start-badges\n\n.. image:: https://github.com/unmade/thrift-pyi/workflows/lint%20and%20test/badge.svg?branch=master\n    :alt: Build Status\n    :target: https://github.com/unmade/thrift-pyi/blob/master/.github/workflows/lint-and-test.yml\n\n.. image:: https://codecov.io/github/unmade/thrift-pyi/coverage.svg?branch=master\n    :alt: Coverage Status\n    :target: https://codecov.io/github/unmade/thrift-pyi\n\n.. image:: https://api.codacy.com/project/badge/Grade/487480f045594e148309e8b7f1f71351\n    :alt: Codacy Badge\n    :target: https://app.codacy.com/app/unmade/thrift-pyi\n\n.. image:: https://requires.io/github/unmade/thrift-pyi/requirements.svg?branch=master\n    :alt: Requirements Status\n    :target: https://requires.io/github/unmade/thrift-pyi/requirements/?branch=master\n\n.. image:: https://img.shields.io/pypi/v/thrift-pyi.svg\n    :alt: PyPI Package latest release\n    :target: https://pypi.org/project/thrift-pyi\n\n.. image:: https://img.shields.io/pypi/wheel/thrift-pyi.svg\n    :alt: PyPI Wheel\n    :target: https://pypi.org/project/thrift-pyi\n\n.. image:: https://img.shields.io/pypi/pyversions/thrift-pyi.svg\n    :alt: Supported versions\n    :target: https://pypi.org/project/thrift-pyi\n\n.. image:: https://img.shields.io/badge/License-MIT-purple.svg\n    :alt: MIT License\n    :target: https://github.com/unmade/thrift-pyi/blob/master/LICENSE\n\n.. end-badges\n\nThis is simple `.pyi` stubs generator from thrift interfaces.\nMotivation for this project is to have autocomplete and type checking\nfor dynamically loaded thrift interfaces\n\nInstallation\n============\n\n.. code-block:: bash\n\n    pip install thrift-pyi\n\nQuickstart\n==========\n\nSample usage:\n\n.. code-block:: bash\n\n    $ thriftpyi example/interfaces --output example/app/interfaces\n\nAdditionally to generated stubs you might want to create `__init__.py` that will load thrift interfaces, for example:\n\n.. code-block:: python\n\n    from pathlib import Path\n    from types import ModuleType\n    from typing import Dict\n\n    import thriftpy2\n\n    _interfaces_path = Path("example/interfaces")\n    _interfaces: Dict[str, ModuleType] = {}\n\n\n    def __getattr__(name):\n        try:\n            return _interfaces[name]\n        except KeyError:\n            interface = thriftpy2.load(str(_interfaces_path.joinpath(f"{name}.thrift")))\n            _interfaces[name] = interface\n            return interface\n\n\nTo see more detailed example of usage refer to `example app <https://github.com/unmade/thrift-pyi/blob/master/example>`_\n\n--strict-optional\n-----------------\n\nPython and thrift are very different at argument handling.\nFor example in thrift the following will be correct declaration:\n\n.. code-block:: thrift\n\n    struct TodoItem {\n        1: required i32 id\n        3: optional i32 type = 1\n        2: required string text\n    }\n\nIn python attributes without a default cannot follow attributes with one.\nTherefore by default all fields are optional with default to None. This is compliant\nto `thriftpy2 <https://github.com/Thriftpy/thriftpy2>`_.\n\nHowever, if you want more strict behaviour you can specify `--strict-optional` option.\nFor the case above, the following stubs will be generated:\n\n.. code-block:: python\n\n    from dataclasses import dataclass\n\n    @dataclass\n    class TodoItem:\n        id: int\n        type: int = 1\n        text: str\n\nDevelopment\n===========\n\nTo install pre-commit hooks::\n\n    pre-commit install\n\nTo run the all tests run::\n\n    tox\n',
    'author': 'Aleksei Maslakov',
    'author_email': 'lesha.maslakov@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/unmade/thrift-pyi',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from thriftpyi.entities import Class, Content, Field, Method, Service
from thriftpyi.proxies import ClassProxy, FieldProxy, InterfaceProxy, ServiceProxy


def make_content(interface: str, is_async: bool, strict_optional: bool) -> Content:
    module = InterfaceProxy(interface)
    return Content(
        imports=list(module.get_imports().keys()),
        errors=[make_class(klass, strict_optional) for klass in module.get_errors()],
        enums=[make_class(klass, strict_optional) for klass in module.get_enums()],
        structs=[make_class(klass, strict_optional) for klass in module.get_structs()],
        services=[make_service(service) for service in module.get_services()],
        is_async=is_async,
    )


def make_class(klass: ClassProxy, strict_optional: bool) -> Class:
    return Class(
        name=klass.name,
        fields=[
            make_field(field, klass.module_name, strict_optional)
            for field in klass.get_fields()
        ],
    )


def make_field(field: FieldProxy, module_name: str, strict_optional: bool) -> Field:
    return Field(
        name=field.name,
        type=field.reveal_type_for(module_name, strict_optional),
        value=field.reveal_value(strict_optional),
    )


def make_service(service: ServiceProxy) -> Service:
    return Service(
        name=service.name,
        methods=[
            make_method(method_name, service) for method_name in service.get_methods()
        ],
    )


def make_method(method_name: str, service: ServiceProxy):
    return Method(
        name=method_name,
        args=[
            Field(name=arg.name, type=arg.reveal_type_for(service.module_name))
            for arg in service.get_args_for(method_name)
        ],
        return_type=service.get_return_type_for(method_name),
    )



import subprocess
from pathlib import Path
from typing import List

from thriftpyi import factories, files, renderers
from thriftpyi.entities import Content


def thriftpyi(
    interfaces_dir: str, output_dir: str, is_async: bool, strict_optional: bool
) -> None:
    interfaces = files.list_interfaces(interfaces_dir)
    _generate_stubs(interfaces, output_dir, is_async, strict_optional)
    _generate_init(interfaces, output_dir)
    path = Path(output_dir).resolve()
    subprocess.check_call([f"autoflake -i -r {path.joinpath('*')}"], shell=True)
    subprocess.check_call(["black", "--quiet", f"{path}"])


def _generate_stubs(
    interfaces: List[str], output_dir: str, is_async: bool, strict_optional: bool
) -> None:
    for interface in interfaces:
        _generate(
            factories.make_content(interface, is_async, strict_optional),
            interface,
            output_dir,
        )


def _generate_init(interfaces: List[str], output_dir: str) -> None:
    imports = sorted([files.get_name(interface) for interface in interfaces])
    _generate(Content(imports=imports), "__init__.pyi", output_dir)


def _generate(content: Content, interface: str, output_dir: str) -> None:
    result = renderers.render(content)
    files.save(result, to=files.build_output_path(interface, output_dir))



import click

from thriftpyi.main import thriftpyi


@click.command()
@click.argument("interfaces_dir", type=click.Path(exists=True))
@click.option("--output", "-o", help="Directory where to save generated `.pyi` files")
@click.option(
    "--async",
    "is_async",
    is_flag=True,
    help="Whether service methods should be async or not",
)
@click.option(
    "--strict-optional",
    is_flag=True,
    help="If not specified all fields will be optional with default to None",
)
def main(interfaces_dir: str, output: str, is_async: bool, strict_optional: bool):
    thriftpyi(interfaces_dir, output, is_async, strict_optional)
